package be.iminds.ilabt.jfed.espec.parser;

import be.iminds.ilabt.jfed.espec.model.DirSpec;
import be.iminds.ilabt.jfed.espec.model.ExperimentSpecification;
import be.iminds.ilabt.jfed.espec.model.FileSource;
import org.testng.annotations.Test;

import static org.testng.Assert.*;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;

import static org.mockito.Mockito.mock;

public class ExperimentSpecificationParserTest {
    @Test
    public void parseA() throws ExperimentSpecificationParser.ExperimentSpecificationParseException {
        String in = "version: 1.0-basic\n" +
                "rspec: nodes.rspec\n" +
                "upload: exp-data-files.tar.gz\n" +
                "execute: exp-script.sh";
        ExperimentSpecificationParser parser =
                new ExperimentSpecificationParser();
        ExperimentSpecification res = parser.parse(in);

        assertThat(res, is(notNullValue()));
        assertThat(res.getVersionBase(), is(equalTo("1.0")));
        assertThat(res.getVersionFlavor(), is(equalTo(ExperimentSpecification.Flavor.BASIC)));
        assertThat(res.getVersionFull(), is(equalToIgnoringCase("1.0-basic")));

        assertThat(res.getRspecs(), hasSize(1));
        assertThat(res.getDirs(), hasSize(2));
        assertThat(res.getUploads(), hasSize(1));
        assertThat(res.getExecutes(), hasSize(1));

        DirSpec a = res.getDirs().get(0);
        DirSpec b = res.getDirs().get(1);

        DirSpec uploadDir, scriptDir;

        if (a.isUploadContent()) {
            uploadDir = a;
            scriptDir = b;
        } else {
            uploadDir = b;
            scriptDir = a;
        }

        assertThat(uploadDir.getDirContent(), is(ESpecConstants.DirContent.UPLOAD));
        assertThat(uploadDir.isUploadContent(), is(true));
        assertThat(uploadDir.getNodes(), is(nullValue()));
        assertThat(uploadDir.getPath(), is(equalTo("~")));
        assertThat(uploadDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(scriptDir.getDirContent(), is(ESpecConstants.DirContent.SCRIPT));
        assertThat(scriptDir.isScriptContent(), is(true));
        assertThat(scriptDir.getNodes(), is(nullValue()));
        assertThat(scriptDir.getPath(), is(equalTo("~")));
        assertThat(scriptDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(res.getRspecs().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getRspecs().get(0).getSource().getValue(), is(equalTo("nodes.rspec")));

        assertThat(res.getExecutes().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getExecutes().get(0).getSource().getValue(), is(equalTo("exp-script.sh")));
        assertThat(res.getExecutes().get(0).getNodes(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPath(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPermissions(), is(equalTo("u=rwx")));
        assertThat(res.getExecutes().get(0).getPwd(), is(nullValue()));

        assertThat(res.getUploads().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getUploads().get(0).getSource().getValue(), is(equalTo("exp-data-files.tar.gz")));
        assertThat(res.getUploads().get(0).getNodes(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPath(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPermissions(), is(equalTo("u=rw")));
    }

    @Test
    public void parseA2() throws ExperimentSpecificationParser.ExperimentSpecificationParseException {
        String in = "version: 1.0-basic\n" +
                "rspec:\n" +
                "   - nodes.rspec\n" +
                "upload:\n" +
                "   - exp-data-files.tar.gz\n" +
                "execute:\n" +
                "   - exp-script.sh\n";
        ExperimentSpecificationParser parser =
                new ExperimentSpecificationParser();
        ExperimentSpecification res = parser.parse(in);

        assertThat(res, is(notNullValue()));
        assertThat(res.getVersionBase(), is(equalTo("1.0")));
        assertThat(res.getVersionFlavor(), is(equalTo(ExperimentSpecification.Flavor.BASIC)));
        assertThat(res.getVersionFull(), is(equalToIgnoringCase("1.0-basic")));

        assertThat(res.getRspecs(), hasSize(1));
        assertThat(res.getDirs(), hasSize(2));
        assertThat(res.getUploads(), hasSize(1));
        assertThat(res.getExecutes(), hasSize(1));

        DirSpec a = res.getDirs().get(0);
        DirSpec b = res.getDirs().get(1);

        DirSpec uploadDir, scriptDir;

        if (a.isUploadContent()) {
            uploadDir = a;
            scriptDir = b;
        } else {
            uploadDir = b;
            scriptDir = a;
        }

        assertThat(uploadDir.getDirContent(), is(ESpecConstants.DirContent.UPLOAD));
        assertThat(uploadDir.isUploadContent(), is(true));
        assertThat(uploadDir.getNodes(), is(nullValue()));
        assertThat(uploadDir.getPath(), is(equalTo("~")));
        assertThat(uploadDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(scriptDir.getDirContent(), is(ESpecConstants.DirContent.SCRIPT));
        assertThat(scriptDir.isScriptContent(), is(true));
        assertThat(scriptDir.getNodes(), is(nullValue()));
        assertThat(scriptDir.getPath(), is(equalTo("~")));
        assertThat(scriptDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(res.getRspecs().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getRspecs().get(0).getSource().getValue(), is(equalTo("nodes.rspec")));

        assertThat(res.getExecutes().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getExecutes().get(0).getSource().getValue(), is(equalTo("exp-script.sh")));
        assertThat(res.getExecutes().get(0).getNodes(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPath(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPermissions(), is(equalTo("u=rwx")));
        assertThat(res.getExecutes().get(0).getPwd(), is(nullValue()));

        assertThat(res.getUploads().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getUploads().get(0).getSource().getValue(), is(equalTo("exp-data-files.tar.gz")));
        assertThat(res.getUploads().get(0).getNodes(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPath(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPermissions(), is(equalTo("u=rw")));
    }
    @Test
    public void parseA2b() throws ExperimentSpecificationParser.ExperimentSpecificationParseException {
        String in = "version: \"1.0-basic\"\n" +
                "rspec: [nodes.rspec]\n" +
                "upload: [ exp-data-files.tar.gz ]\n" +
                "execute: [ \"exp-script.sh\" ]";
        ExperimentSpecificationParser parser =
                new ExperimentSpecificationParser();
        ExperimentSpecification res = parser.parse(in);

        assertThat(res, is(notNullValue()));
        assertThat(res.getVersionBase(), is(equalTo("1.0")));
        assertThat(res.getVersionFlavor(), is(equalTo(ExperimentSpecification.Flavor.BASIC)));
        assertThat(res.getVersionFull(), is(equalToIgnoringCase("1.0-basic")));

        assertThat(res.getRspecs(), hasSize(1));
        assertThat(res.getDirs(), hasSize(2));
        assertThat(res.getUploads(), hasSize(1));
        assertThat(res.getExecutes(), hasSize(1));

        DirSpec a = res.getDirs().get(0);
        DirSpec b = res.getDirs().get(1);

        DirSpec uploadDir, scriptDir;

        if (a.isUploadContent()) {
            uploadDir = a;
            scriptDir = b;
        } else {
            uploadDir = b;
            scriptDir = a;
        }

        assertThat(uploadDir.getDirContent(), is(ESpecConstants.DirContent.UPLOAD));
        assertThat(uploadDir.isUploadContent(), is(true));
        assertThat(uploadDir.getNodes(), is(nullValue()));
        assertThat(uploadDir.getPath(), is(equalTo("~")));
        assertThat(uploadDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(scriptDir.getDirContent(), is(ESpecConstants.DirContent.SCRIPT));
        assertThat(scriptDir.isScriptContent(), is(true));
        assertThat(scriptDir.getNodes(), is(nullValue()));
        assertThat(scriptDir.getPath(), is(equalTo("~")));
        assertThat(scriptDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(res.getRspecs().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getRspecs().get(0).getSource().getValue(), is(equalTo("nodes.rspec")));

        assertThat(res.getExecutes().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getExecutes().get(0).getSource().getValue(), is(equalTo("exp-script.sh")));
        assertThat(res.getExecutes().get(0).getNodes(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPath(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPermissions(), is(equalTo("u=rwx")));
        assertThat(res.getExecutes().get(0).getPwd(), is(nullValue()));

        assertThat(res.getUploads().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getUploads().get(0).getSource().getValue(), is(equalTo("exp-data-files.tar.gz")));
        assertThat(res.getUploads().get(0).getNodes(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPath(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPermissions(), is(equalTo("u=rw")));
    }

    @Test
    public void parseA3() throws ExperimentSpecificationParser.ExperimentSpecificationParseException {
        String in = "version: 1.0-basic\n" +
                "rspec:\n" +
                "   - bundled: nodes.rspec\n" +
                "upload:\n" +
                "   - bundled: exp-data-files.tar.gz\n" +
                "execute:\n" +
                "   - bundled: exp-script.sh";
        ExperimentSpecificationParser parser =
                new ExperimentSpecificationParser();
        ExperimentSpecification res = parser.parse(in);

        assertThat(res, is(notNullValue()));
        assertThat(res.getVersionBase(), is(equalTo("1.0")));
        assertThat(res.getVersionFlavor(), is(equalTo(ExperimentSpecification.Flavor.BASIC)));
        assertThat(res.getVersionFull(), is(equalToIgnoringCase("1.0-basic")));

        assertThat(res.getRspecs(), hasSize(1));
        assertThat(res.getDirs(), hasSize(2));
        assertThat(res.getUploads(), hasSize(1));
        assertThat(res.getExecutes(), hasSize(1));

        DirSpec a = res.getDirs().get(0);
        DirSpec b = res.getDirs().get(1);

        DirSpec uploadDir, scriptDir;

        if (a.isUploadContent()) {
            uploadDir = a;
            scriptDir = b;
        } else {
            uploadDir = b;
            scriptDir = a;
        }

        assertThat(uploadDir.getDirContent(), is(ESpecConstants.DirContent.UPLOAD));
        assertThat(uploadDir.isUploadContent(), is(true));
        assertThat(uploadDir.getNodes(), is(nullValue()));
        assertThat(uploadDir.getPath(), is(equalTo("~")));
        assertThat(uploadDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(scriptDir.getDirContent(), is(ESpecConstants.DirContent.SCRIPT));
        assertThat(scriptDir.isScriptContent(), is(true));
        assertThat(scriptDir.getNodes(), is(nullValue()));
        assertThat(scriptDir.getPath(), is(equalTo("~")));
        assertThat(scriptDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(res.getRspecs().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getRspecs().get(0).getSource().getValue(), is(equalTo("nodes.rspec")));

        assertThat(res.getExecutes().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getExecutes().get(0).getSource().getValue(), is(equalTo("exp-script.sh")));
        assertThat(res.getExecutes().get(0).getNodes(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPath(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPermissions(), is(equalTo("u=rwx")));
        assertThat(res.getExecutes().get(0).getPwd(), is(nullValue()));

        assertThat(res.getUploads().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getUploads().get(0).getSource().getValue(), is(equalTo("exp-data-files.tar.gz")));
        assertThat(res.getUploads().get(0).getNodes(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPath(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPermissions(), is(equalTo("u=rw")));
    }
    @Test
    public void parseA3b() throws ExperimentSpecificationParser.ExperimentSpecificationParseException {
        String in = "version: 1.0-basic\n" +
                "rspec: { bundled: nodes.rspec }\n" +
                "upload: { \"bundled\": \"exp-data-files.tar.gz\" }\n" +
                "execute: { bundled: exp-script.sh }";
        ExperimentSpecificationParser parser =
                new ExperimentSpecificationParser();
        ExperimentSpecification res = parser.parse(in);

        assertThat(res, is(notNullValue()));
        assertThat(res.getVersionBase(), is(equalTo("1.0")));
        assertThat(res.getVersionFlavor(), is(equalTo(ExperimentSpecification.Flavor.BASIC)));
        assertThat(res.getVersionFull(), is(equalToIgnoringCase("1.0-basic")));

        assertThat(res.getRspecs(), hasSize(1));
        assertThat(res.getDirs(), hasSize(2));
        assertThat(res.getUploads(), hasSize(1));
        assertThat(res.getExecutes(), hasSize(1));

        DirSpec a = res.getDirs().get(0);
        DirSpec b = res.getDirs().get(1);

        DirSpec uploadDir, scriptDir;

        if (a.isUploadContent()) {
            uploadDir = a;
            scriptDir = b;
        } else {
            uploadDir = b;
            scriptDir = a;
        }

        assertThat(uploadDir.getDirContent(), is(ESpecConstants.DirContent.UPLOAD));
        assertThat(uploadDir.isUploadContent(), is(true));
        assertThat(uploadDir.getNodes(), is(nullValue()));
        assertThat(uploadDir.getPath(), is(equalTo("~")));
        assertThat(uploadDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(scriptDir.getDirContent(), is(ESpecConstants.DirContent.SCRIPT));
        assertThat(scriptDir.isScriptContent(), is(true));
        assertThat(scriptDir.getNodes(), is(nullValue()));
        assertThat(scriptDir.getPath(), is(equalTo("~")));
        assertThat(scriptDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(res.getRspecs().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getRspecs().get(0).getSource().getValue(), is(equalTo("nodes.rspec")));

        assertThat(res.getExecutes().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getExecutes().get(0).getSource().getValue(), is(equalTo("exp-script.sh")));
        assertThat(res.getExecutes().get(0).getNodes(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPath(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPermissions(), is(equalTo("u=rwx")));
        assertThat(res.getExecutes().get(0).getPwd(), is(nullValue()));

        assertThat(res.getUploads().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getUploads().get(0).getSource().getValue(), is(equalTo("exp-data-files.tar.gz")));
        assertThat(res.getUploads().get(0).getNodes(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPath(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPermissions(), is(equalTo("u=rw")));
    }
    @Test
    public void parseA3c() throws ExperimentSpecificationParser.ExperimentSpecificationParseException {
        String in = "version: 1.0-basic\n" +
                "rspec: [{ bundled: nodes.rspec }]\n" +
                "upload: [{ \"bundled\": \"exp-data-files.tar.gz\" }]\n" +
                "execute: [{ bundled: exp-script.sh }]";
        ExperimentSpecificationParser parser =
                new ExperimentSpecificationParser();
        ExperimentSpecification res = parser.parse(in);

        assertThat(res, is(notNullValue()));
        assertThat(res.getVersionBase(), is(equalTo("1.0")));
        assertThat(res.getVersionFlavor(), is(equalTo(ExperimentSpecification.Flavor.BASIC)));
        assertThat(res.getVersionFull(), is(equalToIgnoringCase("1.0-basic")));

        assertThat(res.getRspecs(), hasSize(1));
        assertThat(res.getDirs(), hasSize(2));
        assertThat(res.getUploads(), hasSize(1));
        assertThat(res.getExecutes(), hasSize(1));

        DirSpec a = res.getDirs().get(0);
        DirSpec b = res.getDirs().get(1);

        DirSpec uploadDir, scriptDir;

        if (a.isUploadContent()) {
            uploadDir = a;
            scriptDir = b;
        } else {
            uploadDir = b;
            scriptDir = a;
        }

        assertThat(uploadDir.getDirContent(), is(ESpecConstants.DirContent.UPLOAD));
        assertThat(uploadDir.isUploadContent(), is(true));
        assertThat(uploadDir.getNodes(), is(nullValue()));
        assertThat(uploadDir.getPath(), is(equalTo("~")));
        assertThat(uploadDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(scriptDir.getDirContent(), is(ESpecConstants.DirContent.SCRIPT));
        assertThat(scriptDir.isScriptContent(), is(true));
        assertThat(scriptDir.getNodes(), is(nullValue()));
        assertThat(scriptDir.getPath(), is(equalTo("~")));
        assertThat(scriptDir.getPermissions(), is(equalTo("u=rwx")));

        assertThat(res.getRspecs().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getRspecs().get(0).getSource().getValue(), is(equalTo("nodes.rspec")));

        assertThat(res.getExecutes().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getExecutes().get(0).getSource().getValue(), is(equalTo("exp-script.sh")));
        assertThat(res.getExecutes().get(0).getNodes(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPath(), is(nullValue()));
        assertThat(res.getExecutes().get(0).getPermissions(), is(equalTo("u=rwx")));
        assertThat(res.getExecutes().get(0).getPwd(), is(nullValue()));

        assertThat(res.getUploads().get(0).getSource().getType(), is(equalTo(FileSource.SourceType.BUNDLED)));
        assertThat(res.getUploads().get(0).getSource().getValue(), is(equalTo("exp-data-files.tar.gz")));
        assertThat(res.getUploads().get(0).getNodes(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPath(), is(nullValue()));
        assertThat(res.getUploads().get(0).getPermissions(), is(equalTo("u=rw")));
    }
}
