package be.iminds.ilabt.jfed.ui.cli2;

import be.iminds.ilabt.jfed.ui.cli2.instruction.*;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.util.Map;

/**
 * InstructionParser
 */
public class InstructionParser {
    private static final Logger LOG = LoggerFactory.getLogger(InstructionParser.class);
    private final static ObjectMapper MAPPER = new ObjectMapper();

    @Nonnull
    public Instruction parse(Instruction.Action action, @Nonnull Map instructionMap) {
        //TODO process the Map into a specific Instruction implementation.
        //This just looks at action, and then uses jackson to translate
        // the instructionMap to an object of the appropriate class
        Class<? extends Instruction> targetClass;
        switch (action) {
            case RUN: {
                targetClass = RunInstruction.class;
                break;
            }
            case CREATESLICE: {
                targetClass = CreateSliceInstruction.class;
                break;
            }
//            case DECRYPTPEM: {
//                targetClass = TODO.class;
//                break;
//            }
            case SLICEINFO: {
                targetClass = SliceInfoInstruction.class;
                break;
            }
            case USERINFO: {
                targetClass = UserInfoInstruction.class;
                break;
            }
            case MANIFEST: {
                targetClass = ManifestInstruction.class;
                break;
            }
            case SLICECREDENTIAL: {
                targetClass = CredentialInstruction.class;
                break;
            }
            case DELETE: {
                targetClass = DeleteInstruction.class;
                break;
            }
            case RENEW: {
                targetClass = RenewInstruction.class;
                break;
            }
            case POA: {
                targetClass = PoaInstruction.class;
                break;
            }
            default:
                throw new RuntimeException("Support for action \""+action+"\" is not impemented.");
        }
//        try {
//            //TODO: jackson might have a direct way to go from MAP to objects. Check the automated tester.
//            String json = MAPPER.writeValueAsString(instructionMap);
//            Instruction res = MAPPER.readValue(json, targetClass);
        String actionInMap = instructionMap.get("action")+"";
        LOG.debug("Trying to parse action "+action+" to "+targetClass+"  (action in map is "+actionInMap+")");
        Instruction res = MAPPER.convertValue(instructionMap, targetClass);
        return res;
//        } catch (IOException e) {
//            throw new RuntimeException("InstructionParser Failed to process JSON instructions", e);
//        }
    }
}
