package be.iminds.ilabt.jfed.experimenter_gui.slice;

import be.iminds.ilabt.jfed.experimenter_gui.canvas.ExperimentCanvas;
import be.iminds.ilabt.jfed.experimenter_gui.util.BrowserUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.ImageUtil;
import be.iminds.ilabt.jfed.rspec.model.WebVncService;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXGeantTestbedType;
import javafx.application.Platform;
import javafx.beans.InvalidationListener;
import javafx.scene.Node;
import javafx.scene.control.Alert;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.FlowPane;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * User: twalcari
 * Date: 10/14/2015
 * Time: 17:13
 */
public class GTSDetailsView extends FlowPane {
    private static final Logger LOG = LoggerFactory.getLogger(GTSDetailsView.class);

    private final BrowserUtil browserUtil;
    private FXGeantTestbedType manifestGeantTestbedType = null;

    private Label noInfoLabel = new Label("No information available");

    {
        setPrefWrapLength(300);
        setHgap(5);
        setVgap(5);
    }

    public GTSDetailsView(BrowserUtil browserUtil) {
        this.browserUtil = browserUtil;

        setManifestGeantTestbedType(null);
    }

    public void setManifestGeantTestbedType(FXGeantTestbedType manifestGeantTestbedType) {
        this.manifestGeantTestbedType = manifestGeantTestbedType;

        List<Node> newChildren = new ArrayList<>();
        if (manifestGeantTestbedType == null) {
            newChildren.add(noInfoLabel);
        } else {
            manifestGeantTestbedType.getWebVncServices().stream()
                    .map(this::createGtsButton)
                    .forEach(newChildren::add);
        }

        getChildren().setAll(newChildren);

    }

    private Button createGtsButton(WebVncService webVnc) {
        Button button = new Button();
        button.getStylesheets().add(ExperimentCanvas.class.getResource("canvas.css").toExternalForm());
        button.getStyleClass().add("canvas-node");


        final ImageView imageView =
                new ImageView(new Image(ImageUtil.class.getResource("/images/gts_types/32/gts-node.png").toExternalForm()));
        imageView.setFitWidth(16);
        imageView.setFitHeight(16);
        button.setGraphic(imageView);
        button.setText(webVnc.getProviderId());

        button.setOnAction(event -> {
            if (webVnc.getVncUrl() != null) {
                browserUtil.openUrlInBrowser(webVnc.getVncUrl());
            } else {
                new Alert(Alert.AlertType.WARNING, "No login info available").show();
            }
        });

        updateGtsStyleClass(button, webVnc.getVncUrl() != null);
        return button;

    }

    private static void updateGtsStyleClass(Button button, boolean hasLoginInfo) {
        //remove all old styleclasses
        button.getStyleClass().removeAll("canvas-node-unknown",
                "canvas-node-ready",
                "canvas-node-changing",
                "canvas-node-unallocated",
                "canvas-node-uninitialised");


        //apply new styleclasses
        button.getStyleClass().add(statusToNodeStyleClass(hasLoginInfo));
    }

    private static String statusToNodeStyleClass(boolean hasLoginInfo) {
        if (hasLoginInfo)
            return "canvas-node-ready";
        else
            return "canvas-node-changing";

    }

}
